package com.fightingchaos.VASSAL.counters;

import java.awt.Point;

import VASSAL.build.GameModule;
import VASSAL.build.module.Map;
import VASSAL.counters.GamePiece;
import VASSAL.counters.Stack;
import VASSAL.tools.SequenceEncoder;


/**
 * Not a real stack but a group of game pieces which is aware of the single piece offsets.
 *
 */
public class Group extends Stack {
	
	public static final String TYPE = "group";
	protected Point[] before = new Point[INCR]; 
	private Point lastPos;

public Group() {
	this(null);
}

public Group(GamePiece p) {
    if (p != null) {
        setMap(p.getMap());
        setPosition(new Point(p.getPosition()));
        add(p);
      }
	setDefaultMetrics(new GroupStackMetrics());
}

/**
 * Returns the stored relative position from the piece to the center of the Group. 
 * In case the piece is not part of the Group will return the position 0,0
 * 
 * @param p	The gamepiece the relative position is asked for
 * @return A new Point object representing the relative position
 */
protected Point getRelativePosition(GamePiece p) {
	int i = indexOf(p);
	if(i >= 0)
		return new Point(before[i]);
	else
		return new Point(0,0);
}

//use this after setting inidividual piece positions so the center of the 
//Stack will be resynchronized
//as boundingBox is expensive you have to call it on your own
protected synchronized void recalculateCenter() {

	if(pieceCount <=0)
		return; 
   	Point summ = new Point(0,0);
   	Point curr = null;
	for (int i = 0; i < pieceCount; i++) {
		curr = getPosition(contents[i]);
		summ.translate(curr.x, curr.y);
	}
	//position group on map
	summ.x /= pieceCount;
	summ.y /= pieceCount;
	int dx = pos.x-summ.x;
	int dy = pos.y-summ.y;
	for (int i = 0; i < pieceCount; i++) {
		before[i].translate(dx, dy);
	}
	setPosition(summ);
}

/**
 * Returns the absolute map position of the passed gamepiece or in case 
 * the piece is not part of the Group the position of the Group
 * 
 * @param p The GamePiece the absolute position is requested for
 * @return A new Point containing the abolute position
 */
public Point getPosition(GamePiece p) {
	Point ret = getRelativePosition(p);
	ret.translate(pos.x, pos.y);
	return ret;
}

//do a recalculateCenter() hereafter, or after all setPosition calls of the stack 
public void setPosition(GamePiece g,Point p) {
	int i = indexOf(g);
	if(i >= 0)
		before[i] = new Point(p.x-pos.x, p.y-pos.y);
}

public void insertChild(GamePiece child, int index) {
    if ( child.getParent() != null) {
    	child.getParent().remove(child);
      child.setParent(null);
    }
    else if (child.getMap() != null) {
      child.getMap().removePiece(child);
    }
    lastPos = child.getPosition();
    child.setParent(this);
    insertPieceAt(child, index);
    lastPos = null;
  }

protected void insertPieceAt(GamePiece p, int index) {
    if (pieceCount >= contents.length) {
      GamePiece[] newContents = new GamePiece[contents.length + INCR];
      System.arraycopy(contents, 0, newContents, 0, pieceCount);
      contents = newContents;
      
      Point[] newPoints = new Point[contents.length + INCR];
      System.arraycopy(before, 0, newPoints, 0, pieceCount);
      before = newPoints;
    }
    for (int i = pieceCount; i > index; --i) {
      contents[i] = contents[i - 1];
      before[i] = before[i - 1];
    }
    contents[index] = p;
    before[index] = new Point(lastPos.x-pos.x,lastPos.y-pos.y);
    lastPos =null;
    pieceCount++;
    if(pieceCount >1)
    	recalculateCenter();
  }

protected void removePieceAt(int index) {
    if (index >= 0 && index < pieceCount) {
      //set pos relative to us otherwise the piece will drop to the center of teh group :(
      contents[index].setPosition(new Point(pos.x +before[index].x, pos.y +before[index].y));
      pieceCount--;
      for (int i = index; i < pieceCount; ++i) {
        contents[i] = contents[i + 1];
        before[i] = before[i + 1];
      }
      setExpanded(isExpanded() && pieceCount > 1);
      recalculateCenter();
    }
  }

public void insert(GamePiece p, int pos) {
    if (p == null) {
      return;
    }
    pos = Math.max(pos, 0);
    pos = Math.min(pos, pieceCount);
    int index = indexOf(p);
    if (index >= 0) {
      lastPos = new Point(before[index].x+getPosition().x, before[index].y+getPosition().y) ;//save las delta pos
      if (pos > index) {
        insertPieceAt(p, pos + 1);
        removePieceAt(index);
      }
      else {
        removePieceAt(index);
        insertPieceAt(p, pos);
      }
    }
    else {
      insertChild(p, pos);
    }
  }

//FIXME this is not pretty, better would be an overridable myGetType and getType only returning this
public String getType() {
    return TYPE;
  }

/**
 * Also output the absolute position of each piece belonging to this group for later
 * reconstruction
 */
public String getState() {
  SequenceEncoder se = new SequenceEncoder(';');
  se.append(getMap() == null ? "null" : getMap().getIdentifier()).append(getPosition().x).append(getPosition().y);
  Point p = null;
  for (int i = 0; i < pieceCount; ++i) {
    se.append(contents[i].getId());
    //use absolute position here
    p = getPosition(contents[i]);
    se.append(p.x).append(p.y);
  }
  return se.getValue();
}
/**
 * The setState also needs the position of each individual piece for later 
 * reconstruction, hence we need an own setState
 */
public void setState(String s) {
  SequenceEncoder.Decoder st = new SequenceEncoder.Decoder(s, ';');
  String mapId = st.nextToken();
  setPosition(new Point(st.nextInt(0), st.nextInt(0)));
  pieceCount = 0;
  Point p = null;
  while (st.hasMoreTokens()) {
    GamePiece child = GameModule.getGameModule().getGameState().getPieceForId(st.nextToken());
    if (child != null) {
	  //will be absolute position
      p = new Point(st.nextInt(0),st.nextInt(0));
      child.setPosition(p);
      insertChild(child, pieceCount);
    }
  }
  Map m = null;
  if (!"null".equals(mapId)) {
    m = Map.getMapById(mapId);
    if (m == null) {
      throw new RuntimeException("Could not find map " + mapId);
    }
  }
  if (m != getMap()) {
    if (m != null) {
      m.addPiece(this);
    }
    else {
      setMap(null);
    }
  }
}

}
